/*======================================================================
  Irrlicht Engine Wrapped
 Copyright (C) 2004 Ole Ciliox (ole@ira.uka.de)

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 cgkit homepage: http://cgkit.sourceforge.net
======================================================================*/

#include "IrrCore.h"

IrrDependent::IrrDependent() {};
IrrDependent::~IrrDependent() {};
IrrCameraDependent::IrrCameraDependent() {};
IrrCameraDependent::~IrrCameraDependent() {};
IrrLightDependent::IrrLightDependent() {};
IrrLightDependent::~IrrLightDependent() {};



void createMesh(support3d::TriMeshGeom *trigeom, SMeshBuffer *pBuffer)
{
	int i,j;
	bool normals;

	cout << "--------------------------------------------------------" << endl;
	cout << "creating manual mesh buffer" << endl;
	if (trigeom == 0)
		return;

	pBuffer->Vertices  = *(new array<S3DVertex>);
	pBuffer->Indices = *(new array<u16>);

	cout << "vertices @ "<< trigeom->verts.dataPtr() << endl;
	cout << "indices @ "<< trigeom->faces.dataPtr() << endl;

	support3d::vec3d* pVertices = trigeom->verts.dataPtr();
	support3d::vec3d* vertsptr2 = trigeom->verts.dataPtr();
	int* faceptr = trigeom->faces.dataPtr();
	int* faceptr2 = trigeom->faces.dataPtr();
	int* faceptr3 = trigeom->faces.dataPtr();

	int numfaces = trigeom->faces.size();
	int numverts = trigeom->verts.size();

	cout << "creating temporary buffers" << endl;

	S3DVertex* pVertexInfo = new S3DVertex[numverts];
	u16* pIndices = new u16[numfaces*3];

	cout << "checking normals" << endl;
	normals = false;

	support3d::PrimVarInfo *pNormalsVarInfo = trigeom->findVariable("N");
	support3d::vec3d *pNormals=0;
	support3d::ArraySlot<support3d::vec3d>* normsptr = 0;

	if (pNormalsVarInfo == 0 
		|| pNormalsVarInfo->storage != support3d::VARYING 
		|| pNormalsVarInfo->type != support3d::NORMAL 
		|| pNormalsVarInfo->multiplicity != 1	)
	{
		cout << "could not assign normals, hence i will calculate them" << endl;
		normals = false;
	}
	else
	{
		cout << "reading normals ..." << endl;

		normsptr = dynamic_cast<support3d::ArraySlot<support3d::vec3d>*>(pNormalsVarInfo->slot);
		pNormals = normsptr->dataPtr();

		cout << "...count: " << normsptr->size() << endl;
		normals = true;
	}
	// ......................... Vorarbeit Ende
	
	support3d::vec3d *normalsVector=0;
	normalsVector = new support3d::vec3d[numverts];

	for (i = 0; i < numfaces; i++)
	{
		int ia,ib,ic;
		support3d::vec3d normal;
		normal.set(0.0, 0.0, 0.0);

		ia = *faceptr;
		pIndices[i*3+0] = *faceptr;
		faceptr++;

		ib = *faceptr;
		pIndices[i*3+1] = *faceptr;
		faceptr++;

		ic = *faceptr;
		pIndices[i*3+2] = *faceptr;
		faceptr++;

		// NORMALS
		if (!normals)
		{
			try
			{
				normal = ( normal.cross( *(vertsptr2+ib) - *(vertsptr2+ia), *(vertsptr2+ic) - *(vertsptr2+ia) ) );
				normal = normal.normalize();
			}
			catch(...)
			{
				cout << "Warning: (0,0,0) normal detected." << endl;
				normal.set(0,0,0);
			}

			// NORMALS part (face-normals) setzen, also insgesamt 3mal, zu den korrespondierenden ia, ib, ic vertices
			normalsVector[ia]=normal;
			normalsVector[ib]=normal;
			normalsVector[ic]=normal;
		}
	}
	cout << "index buffer filled" << endl;

	for (i = 0; i < numverts; i++) 
	{
		if (normals)
		{
			pVertexInfo[i] = (	S3DVertex(  pVertices->x, pVertices->y, pVertices->z,  
											pNormals->x, pNormals->y, pNormals->z, 
											SColor(128, 255, 128, 128), 
											0.0f, 0.0f )	);
		//	cout << "normals (preset): " << pNormals->x << " " << pNormals->y << " " << pNormals->z << " " << endl;
		}
		else
		{
			pVertexInfo[i] = (	S3DVertex(  pVertices->x, pVertices->y, pVertices->z,  
											normalsVector[i].x, normalsVector[i].y, normalsVector[i].z, 
											SColor(128, 255, 128, 128), 
											0.0f, 0.0f )	);
		//	cout << "normals (preset): " << normalsVector[i].x << " " << normalsVector[i].y << " " << normalsVector[i].z << " " << endl;
		}
		pVertices++;
		pNormals++;
	}
	cout << "vertex buffer filled" << endl;

	pBuffer->Vertices.set_pointer(pVertexInfo, numverts);
	cout << "should point to: " << pVertexInfo << endl;
	cout << "but points to: " << pBuffer->Vertices.pointer() << endl;
	pBuffer->Indices.set_pointer(pIndices, numfaces*3);

	cout << "vertex count: " << pBuffer->getVertexCount() << endl;
	cout << "index count: " << pBuffer->getIndexCount() << endl;
	cout << "--------------------------------------------------------" << endl;
}

void IrrDependent::setup(support3d::WorldObject *pWorldObject, ISceneNode *pSceneNode)
{
	mpWorldObject = pWorldObject;
	mpSceneNode = pSceneNode;
}

void IrrDependent::onValueChanged()
{
	support3d::vec3d changePos, changeScale;
	support3d::mat4d changeRot, changeTransformation;
	vector3df eulerVector, positionVector, scaleVector;

	// get local transformation of the node
	changeTransformation = mpWorldObject->localTransform();
	changeTransformation.decompose(changePos, changeRot, changeScale);
	positionVector.X = changePos[0]; positionVector.Y = changePos[1]; positionVector.Z = changePos[2];
	scaleVector.X = changeScale[0]; scaleVector.Y = changeScale[1]; scaleVector.Z = changeScale[2]; 

	// make euler winkel ----------------------------->
    f64 Y = -asin(changeRot.at(2,0)); 
    f64 D = Y; 
    f64 C = cos(Y); 
    Y *= GRAD_PI; 
    f64 rotx, roty, X, Z; 
    if (fabs(C)>0.0005f)  
	{ 
        rotx = changeRot.at(2,2) / C; 
        roty = changeRot.at(2,1)  / C; 
	    X = atan2( roty, rotx ) * GRAD_PI; 
        rotx = changeRot.at(0,0) / C; 
        roty = changeRot.at(1,0) / C; 
        Z = atan2( roty, rotx ) * GRAD_PI; 
    } 
    else 
    { 
        X  = 0.0f; 
        rotx = changeRot.at(1,1); 
        roty = -changeRot.at(0,1); 
        Z  = atan2( roty, rotx ) * (f32)GRAD_PI; 
    } 
    if (X < 0.00) X += 360.00; 
    if (Y < 0.00) Y += 360.00; 
    if (Z < 0.00) Z += 360.00; 
    eulerVector.X = (f32)X;
	eulerVector.Y = (f32)Y;
	eulerVector.Z = (f32)Z;
	// <-------------------- finished making euler winkel
	
	// set new transformation
	mpSceneNode->setPosition( positionVector );
	mpSceneNode->setRotation( eulerVector );
	mpSceneNode->setScale( scaleVector );
}

void IrrLightDependent::onValueChanged()
{
	cout << "light value changed!" << endl;

	support3d::vec3d changePos, changeScale;
	support3d::mat4d changeRot, changeTransformation;
	vector3df eulerVector, positionVector, scaleVector;

	// get local transformation of the node
	changeTransformation = mpWorldObject->localTransform();
	changeTransformation.decompose(changePos, changeRot, changeScale);
	positionVector.X = changePos[0]; positionVector.Y = changePos[1]; positionVector.Z = changePos[2];

	// make euler winkel ----------------------------->
    f64 Y = -asin(changeRot.at(2,0)); 
    f64 D = Y; 
    f64 C = cos(Y); 
    Y *= GRAD_PI; 
    f64 rotx, roty, X, Z; 
    if (fabs(C)>0.0005f)  
	{ 
        rotx = changeRot.at(2,2) / C; 
        roty = changeRot.at(2,1)  / C; 
	    X = atan2( roty, rotx ) * GRAD_PI; 
        rotx = changeRot.at(0,0) / C; 
        roty = changeRot.at(1,0) / C; 
        Z = atan2( roty, rotx ) * GRAD_PI; 
    } 
    else 
    { 
        X  = 0.0f; 
        rotx = changeRot.at(1,1); 
        roty = -changeRot.at(0,1); 
        Z  = atan2( roty, rotx ) * (f32)GRAD_PI; 
    } 
    if (X < 0.00) X += 360.00; 
    if (Y < 0.00) Y += 360.00; 
    if (Z < 0.00) Z += 360.00; 
    eulerVector.X = (f32)X;
	eulerVector.Y = (f32)Y;
	eulerVector.Z = (f32)Z;
	// <-------------------- finished making euler winkel
	
	// set new transformation
	mpSceneNode->setPosition( positionVector );
	mpSceneNode->setRotation( eulerVector );
}

void IrrCameraDependent::onValueChanged()
{
	support3d::vec3d changePos, changeScale;
	support3d::mat4d changeRot, changeTransformation;
	vector3df eulerVector, positionVector;

	cout << "camera position changed!" << endl;

	// get local transformation of the node
	changeTransformation = mpWorldObject->localTransform();
	changeTransformation.decompose(changePos, changeRot, changeScale);
	positionVector.X = changePos[0]; positionVector.Y = changePos[1]; positionVector.Z = changePos[2];

	// make euler winkel ----------------------------->
    f64 Y = -asin(changeRot.at(2,0)); 
    f64 D = Y; 
    f64 C = cos(Y); 
    Y *= GRAD_PI; 
    f64 rotx, roty, X, Z; 
    if (fabs(C)>0.0005f)  
	{ 
        rotx = changeRot.at(2,2) / C; 
        roty = changeRot.at(2,1)  / C; 
	    X = atan2( roty, rotx ) * GRAD_PI; 
        rotx = changeRot.at(0,0) / C; 
        roty = changeRot.at(1,0) / C; 
        Z = atan2( roty, rotx ) * GRAD_PI; 
    } 
    else 
    { 
        X  = 0.0f; 
        rotx = changeRot.at(1,1); 
        roty = -changeRot.at(0,1); 
        Z  = atan2( roty, rotx ) * (f32)GRAD_PI; 
    } 
    if (X < 0.00) X += 360.00; 
    if (Y < 0.00) Y += 360.00; 
    if (Z < 0.00) Z += 360.00; 
    eulerVector.X = (f32)X;
	eulerVector.Y = (f32)Y;
	eulerVector.Z = (f32)Z;
	// <-------------------- finished making euler winkel
	
	// set new transformation
	//mpSceneNode->setPosition( positionVector );
	//mpSceneNode->setRotation( eulerVector );
	
}

void IrrCore::foo()
{
}

IrrCore::IrrCore() {};
IrrCore::~IrrCore() {};

void IrrCore::setCameraStuff(support3d::WorldObject& camera, float nearclip, float farclip, float fov)
{
	/*
	ISceneNode* pCamNode = mpSmgr->addCameraSceneNode(mpSmgr->getRootSceneNode(), vector3df(0,30,-40), vector3df(0,5,0));
	IrrDependent* pCamDep = new IrrCameraDependent();
	pCamDep->setup(&camera, pCamNode); 
	(camera.transform).addDependent(pCamDep);
	mDependents.push_back(pCamDep);
	*/
  	
	mpCamNode = mpSmgr->addCameraSceneNodeFPS(0,100.0f,300.0f);
	mpCamNode->setPosition(core::vector3df(-3,3,-3));

	// disable mouse cursor
	mpDevice->getCursorControl()->setVisible(false);

	//mpCamNode = mpSmgr->addCameraSceneNodeFPS(0, 100, 500, -1, keyMap, 8);

}

void IrrCore::buildTree(support3d::WorldObject& visumRoot, ISceneNode *pSnode)
{
	support3d::WorldObject::ChildIterator childit = visumRoot.childsBegin();

	// Recursive exit condition: there are no children
	if (childit->second.get() == 0)
	{
		cout << "no childs left!" << endl;
		return;
	}	

	cout << "node found" << endl;

	while ( childit != visumRoot.childsEnd() )
	{
		// Some needed pointers
		ISceneNode *pSceneNode = 0;

		support3d::LightSource *p_ls = 0;
		p_ls = dynamic_cast<support3d::LightSource*>( childit->second.get() );

		// ----------------------------
		// ----------------------------
		// geometry found? ------------
		// ----------------------------
		// ----------------------------
		if ((*childit).second->getGeom().get() != 0)
		{
			cout << "geometry found" << endl;
			SMesh* pMesh = new SMesh();
			SMeshBuffer *pBuffer = new SMeshBuffer();
			pMesh->addMeshBuffer(pBuffer);
			
			support3d::TriMeshGeom *pTriGeom = 0;
			pTriGeom = dynamic_cast<support3d::TriMeshGeom*>( (*childit).second->getGeom().get() );
			if ( pTriGeom != 0 )
			{
				createMesh(pTriGeom, pBuffer);
			}
			else
			{
				support3d::TriMeshGeom *trim = 0;
				trim = new support3d::TriMeshGeom();			

				try
				{
					(childit->second->getGeom().get())->convert(trim);
					// Create an ogre mesh based on the TriMesh
					createMesh(trim, pBuffer);	
				}
				catch(...) // Could not convert
				{
					cout << "error error error error" << endl;
				}
			}
			pSceneNode = mpSmgr->addMeshSceneNode(pMesh, pSnode);
			pSceneNode->setAutomaticCulling(false);

			IrrDependent *pTemp = new IrrDependent();
			pTemp->setup( (*childit).second.get(), pSceneNode );

			// Make it transform dependent and visible dependent
			((*childit).second->transform).addDependent(pTemp);
			((*childit).second->visible).addDependent(pTemp);

			// Finally, store it in our dependents vector
			mDependents.push_back(pTemp);
		}

		// ----------------------------
		// ----------------------------
		// lightsource found? ---------
		// ----------------------------
		// ----------------------------
		else if(p_ls!=0)
		{
			cout << "LLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLL" << endl;
			cout << "light node found!" << endl;
			// supported light types
			support3d::LightSource* tmpSource = 0;
			support3d::GLDistantLight* tmpDist = 0;
			support3d::GLPointLight* tmpPoint = 0;
			support3d::GLSpotLight* tmpSpot = 0;

			// check for light type
			tmpDist = dynamic_cast<support3d::GLDistantLight*>( childit->second.get());
			tmpPoint = dynamic_cast<support3d::GLPointLight*>( childit->second.get());
			tmpSpot = dynamic_cast<support3d::GLSpotLight*>( childit->second.get());

			// safe, as you already are in correct if case
			tmpSource = static_cast<support3d::LightSource*>( childit->second.get());

			pSceneNode = mpSmgr->addLightSceneNode(pSnode, vector3df(0,0,0), SColorf(0.5f, 0.5f, 0.0f));

			cout << "creating dynamic light" << endl;
			IrrDependent* temp_dependent = new IrrLightDependent();
			temp_dependent->setup(tmpSource, pSceneNode);

			((*childit).second->transform).addDependent(temp_dependent);
			mDependents.push_back(temp_dependent);

			/*
			// create a standard transformation dependency
			((*childit).second->transform).addDependent(testdependent);			
			tmpSource->enabled.addDependent(testdependent);	
			tmpSource->intensity.addDependent(testdependent);

			// shadow dependency
			tmpSource->cast_shadow.addDependent(testdependent);
			*/
			cout << "LLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLL" << endl;
		}

		// Recursion: do the same with each child node
		buildTree( *(childit->second), pSceneNode);

		// Back from recursion
		childit++;
	} // end while 
} 

void IrrCore::testSetup(support3d::WorldObject& visumRoot)
{
	playControl = true;
	CgkitEventReceiver *pEvrc= new CgkitEventReceiver(this);
	cout << "created event receiver @ " << pEvrc << endl;
	mpDevice = createDevice(EDT_OPENGL, dimension2d<s32>(640, 480), 16, false, false, false, pEvrc);
	mpDriver = mpDevice->getVideoDriver();
	mpSmgr = mpDevice->getSceneManager();

	// cam stuff
	mpCamNode = mpSmgr->addCameraSceneNodeFPS(0,90.0f,10.0f);
	mpCamNode->setPosition(core::vector3df(-2,2,-2));
	// disable mouse cursor
	mpDevice->getCursorControl()->setVisible(false);
 
	mpDevice->setWindowCaption(L"ein kleiner test");

	buildTree(visumRoot, mpSmgr->getRootSceneNode());
	/*
	while(playControl)
	{
			mpDevice->run();
			mpDriver->beginScene(true, true, SColor(0,200,200,200));
			mpSmgr->drawAll();
			mpDriver->endScene();
	}
	mpDevice->drop();
	*/
}

bool IrrCore::renderTestSetup()
{
	if(playControl)
	{
			mpDevice->run();
			mpDriver->beginScene(true, true, SColor(0,200,200,200));
			mpSmgr->drawAll();
			mpDriver->endScene();
			return true;
	}
	else
	{
		mpDevice->drop();
		return false; 
	}
}


