/*======================================================================
  Irrlicht Engine Wrapped
 Copyright (C) 2004 Ole Ciliox (ole@ira.uka.de)

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 cgkit homepage: http://cgkit.sourceforge.net
======================================================================*/
 
#ifndef __IrrCore_H__
#define __IrrCore_H__

#if OGRE_PLATFORM == PLATFORM_WIN32
#define WIN32_LEAN_AND_MEAN
#include "windows.h"
#endif
#undef max
#undef min
#include "worldobject.h"
#include "geomobject.h"
#include "trimeshgeom.h"
#include "lightsource.h"
#include "gldistantlight.h"
#include "glpointlight.h"
#include "glspotlight.h"
#include <iostream>
#include "slot.h"
#include "quat.h"
#include "stdio.h"
#include "fstream.h"
#include "boost/python/list.hpp"
#include "boost/python/tuple.hpp"

#include "irrlicht.h"

using std::cout;
using std::endl;

using namespace irr;
using namespace core;
using namespace scene;
using namespace video;
using namespace io;
using namespace gui;

/**/

class IrrDependent : public support3d::Dependent
{
public:
	ISceneNode *mpSceneNode;
	support3d::WorldObject* mpWorldObject;

	virtual void onValueChanged();
	virtual void setup(support3d::WorldObject* pWorldObject, ISceneNode *pSceneNode);
	IrrDependent();
	~IrrDependent();
};

class IrrLightDependent : public IrrDependent
{
public:
	IrrLightDependent();
	~IrrLightDependent();
	ILightSceneNode *mpLightSceneNode;
	short lightType; // 0 point, 1 distant, 2 spot
	virtual void onValueChanged();	
};

class IrrCameraDependent : public IrrDependent
{
public:
	IrrCameraDependent();
	~IrrCameraDependent();
	virtual void onValueChanged();	
};

class IrrMeshDependent : public IrrDependent
{
};

class IrrCore
{
public:
	IrrCore();
	~IrrCore();
	void foo();

	bool playControl;

	void testSetup(support3d::WorldObject& visumRoot);
	bool renderTestSetup();

	void setCameraStuff(support3d::WorldObject& camera, float nearclip, float farclip, float fov);
	void buildTree(support3d::WorldObject& visumRoot, ISceneNode *pSnode);

	std::vector<IrrDependent*> mDependents;

private:
	IrrlichtDevice *mpDevice;
	IVideoDriver* mpDriver;
	ISceneManager* mpSmgr;
	IGUIEnvironment* mpGuiEnv;
	SMesh* mpTestMesh;

	ISceneNode *mpCamNode;
};

class CgkitEventReceiver : public IEventReceiver
{
public:
	IrrCore *mpCore;
	CgkitEventReceiver(IrrCore *pCore)
	{
		// constructor
		mpCore = pCore;
	}
	virtual bool OnEvent(SEvent event)
	{

		if (event.EventType == irr::EET_KEY_INPUT_EVENT	&& !event.KeyInput.PressedDown)
		{
			switch(event.KeyInput.Key)
			{
			case KEY_KEY_Q:
				{
					cout << "pressed Q!" << endl;
					mpCore->playControl = false;

				}
			case KEY_KEY_S:
				{
					/*
					core::vector3df v = node->getPosition();
					v.Y += event.KeyInput.Key == KEY_KEY_W ? 2.0f : -2.0f;
					node->setPosition(v);
					*/
					cout << "pressed W or S" << endl;
				}
				return true;
			}
		}

		return false;
	}
};

#endif
